@@ -30,42 +30,22 @@ import sqlite3
 import os
 import sys
 import tempfile
-import random
 import logging
+import time
 import traceback
 
 from threading import Thread
 
-__version__ = '1.7.0'
+__version__ = '2.0.0'
 
-major_version = sys.version_info[0]
-if major_version < 3:  # py <= 2.x
-    if sys.version_info[1] < 5:  # py <= 2.4
-        raise ImportError("sqlitedict requires python 2.5 or higher (python 3.3 or higher supported)")
 
-    # necessary to use exec()_ as this would be a SyntaxError in python3.
-    # this is an exact port of six.reraise():
-    def exec_(_code_, _globs_=None, _locs_=None):
-        """Execute code in a namespace."""
-        if _globs_ is None:
-            frame = sys._getframe(1)
-            _globs_ = frame.f_globals
-            if _locs_ is None:
-                _locs_ = frame.f_locals
-            del frame
-        elif _locs_ is None:
-            _locs_ = _globs_
-        exec("""exec _code_ in _globs_, _locs_""")
+def reraise(tp, value, tb=None):
+    if value is None:
+        value = tp()
+    if value.__traceback__ is not tb:
+        raise value.with_traceback(tb)
+    raise value
 
-    exec_("def reraise(tp, value, tb=None):\n"
-          "    raise tp, value, tb\n")
-else:
-    def reraise(tp, value, tb=None):
-        if value is None:
-            value = tp()
-        if value.__traceback__ is not tb:
-            raise value.with_traceback(tb)
-        raise value
 
 try:
     from cPickle import dumps, loads, HIGHEST_PROTOCOL as PICKLE_PROTOCOL
@@ -106,7 +86,8 @@ class SqliteDict(DictClass):
     VALID_FLAGS = ['c', 'r', 'w', 'n']
 
     def __init__(self, filename=None, tablename='unnamed', flag='c',
-                 autocommit=False, journal_mode="DELETE", encode=encode, decode=decode):
+                 autocommit=False, journal_mode="DELETE", encode=encode,
+                 decode=decode, timeout=5, outer_stack=True):
         """
         Initialize a thread-safe sqlite-backed dictionary. The dictionary will
         be a table `tablename` in database file `filename`. A single file (=database)
@@ -122,6 +103,10 @@ class SqliteDict(DictClass):
         Set `journal_mode` to 'OFF' if you're experiencing sqlite I/O problems
         or if you need performance and don't care about crash-consistency.
 
+        Set `outer_stack` to False to disable the output of the outer exception
+        to the error logs.  This may improve the efficiency of sqlitedict
+        operation at the expense of a detailed exception trace.
+
         The `flag` parameter. Exactly one of:
           'c': default mode, open for read/write, creating the db/table if necessary.
           'w': open for r/w, but drop `tablename` contents first (start with empty table)
@@ -137,11 +122,13 @@ class SqliteDict(DictClass):
         object.
         The default is to use pickle.
 
+        The `timeout` defines the maximum time (in seconds) to wait for initial Thread startup.
+
         """
         self.in_temp = filename is None
         if self.in_temp:
-            randpart = hex(random.randint(0, 0xffffff))[2:]
-            filename = os.path.join(tempfile.gettempdir(), 'sqldict' + randpart)
+            fd, filename = tempfile.mkstemp(prefix='sqldict')
+            os.close(fd)
 
         if flag not in SqliteDict.VALID_FLAGS:
             raise RuntimeError("Unrecognized flag: %s" % flag)
@@ -166,17 +153,30 @@ class SqliteDict(DictClass):
         self.journal_mode = journal_mode
         self.encode = encode
         self.decode = decode
+        self.timeout = timeout
+        self._outer_stack = outer_stack
 
         logger.info("opening Sqlite table %r in %r" % (tablename, filename))
-        MAKE_TABLE = 'CREATE TABLE IF NOT EXISTS "%s" (key TEXT PRIMARY KEY, value BLOB)' % self.tablename
         self.conn = self._new_conn()
-        self.conn.execute(MAKE_TABLE)
-        self.conn.commit()
+        if self.flag == 'r':
+            if self.tablename not in SqliteDict.get_tablenames(self.filename):
+                msg = 'Refusing to create a new table "%s" in read-only DB mode' % tablename
+                raise RuntimeError(msg)
+        else:
+            MAKE_TABLE = 'CREATE TABLE IF NOT EXISTS "%s" (key TEXT PRIMARY KEY, value BLOB)' % self.tablename
+            self.conn.execute(MAKE_TABLE)
+            self.conn.commit()
         if flag == 'w':
             self.clear()
 
     def _new_conn(self):
-        return SqliteMultithread(self.filename, autocommit=self.autocommit, journal_mode=self.journal_mode)
+        return SqliteMultithread(
+            self.filename,
+            autocommit=self.autocommit,
+            journal_mode=self.journal_mode,
+            timeout=self.timeout,
+            outer_stack=self._outer_stack,
+        )
 
     def __enter__(self):
         if not hasattr(self, 'conn') or self.conn is None:
@@ -225,13 +225,13 @@ class SqliteDict(DictClass):
             yield key, self.decode(value)
 
     def keys(self):
-        return self.iterkeys() if major_version > 2 else list(self.iterkeys())
+        return self.iterkeys()
 
     def values(self):
-        return self.itervalues() if major_version > 2 else list(self.itervalues())
+        return self.itervalues()
 
     def items(self):
-        return self.iteritems() if major_version > 2 else list(self.iteritems())
+        return self.iteritems()
 
     def __contains__(self, key):
         HAS_ITEM = 'SELECT 1 FROM "%s" WHERE key = ?' % self.tablename
@@ -288,7 +288,8 @@ class SqliteDict(DictClass):
         if self.flag == 'r':
             raise RuntimeError('Refusing to clear read-only SqliteDict')
 
-        CLEAR_ALL = 'DELETE FROM "%s";' % self.tablename  # avoid VACUUM, as it gives "OperationalError: database schema has changed"
+        # avoid VACUUM, as it gives "OperationalError: database schema has changed"
+        CLEAR_ALL = 'DELETE FROM "%s";' % self.tablename
         self.conn.commit()
         self.conn.execute(CLEAR_ALL)
         self.conn.commit()
@@ -331,7 +332,7 @@ class SqliteDict(DictClass):
         if self.in_temp:
             try:
                 os.remove(self.filename)
-            except:
+            except Exception:
                 pass
 
     def terminate(self):
@@ -361,12 +362,6 @@ class SqliteDict(DictClass):
             # in __del__ method.
             pass
 
-# Adding extra methods for python 2 compatibility (at import time)
-if major_version == 2:
-    SqliteDict.__nonzero__ = SqliteDict.__bool__
-    del SqliteDict.__bool__  # not needed and confusing
-#endclass SqliteDict
-
 
 class SqliteMultithread(Thread):
     """
@@ -376,7 +371,7 @@ class SqliteMultithread(Thread):
     in a separate thread (in the same order they arrived).
 
     """
-    def __init__(self, filename, autocommit, journal_mode):
+    def __init__(self, filename, autocommit, journal_mode, timeout, outer_stack=True):
         super(SqliteMultithread, self).__init__()
         self.filename = filename
         self.autocommit = autocommit
@@ -385,19 +380,35 @@ class SqliteMultithread(Thread):
         self.reqs = Queue()
         self.setDaemon(True)  # python2.5-compatible
         self.exception = None
+        self._sqlitedict_thread_initialized = None
+        self._outer_stack = outer_stack
+        self.timeout = timeout
         self.log = logging.getLogger('sqlitedict.SqliteMultithread')
         self.start()
 
     def run(self):
-        if self.autocommit:
-            conn = sqlite3.connect(self.filename, isolation_level=None, check_same_thread=False)
-        else:
-            conn = sqlite3.connect(self.filename, check_same_thread=False)
-        conn.execute('PRAGMA journal_mode = %s' % self.journal_mode)
-        conn.text_factory = str
-        cursor = conn.cursor()
-        conn.commit()
-        cursor.execute('PRAGMA synchronous=OFF')
+        try:
+            if self.autocommit:
+                conn = sqlite3.connect(self.filename, isolation_level=None, check_same_thread=False)
+            else:
+                conn = sqlite3.connect(self.filename, check_same_thread=False)
+        except Exception:
+            self.log.exception("Failed to initialize connection for filename: %s" % self.filename)
+            self.exception = sys.exc_info()
+            raise
+
+        try:
+            conn.execute('PRAGMA journal_mode = %s' % self.journal_mode)
+            conn.text_factory = str
+            cursor = conn.cursor()
+            conn.commit()
+            cursor.execute('PRAGMA synchronous=OFF')
+        except Exception:
+            self.log.exception("Failed to execute PRAGMA statements.")
+            self.exception = sys.exc_info()
+            raise
+
+        self._sqlitedict_thread_initialized = True
 
         res = None
         while True:
@@ -412,7 +423,7 @@ class SqliteMultithread(Thread):
             else:
                 try:
                     cursor.execute(req, arg)
-                except Exception as err:
+                except Exception:
                     self.exception = (e_type, e_value, e_tb) = sys.exc_info()
                     inner_stack = traceback.extract_stack()
 
@@ -434,10 +445,18 @@ class SqliteMultithread(Thread):
                         self.log.error(item)
 
                     self.log.error('')  # exception & outer stack w/blank line
-                    self.log.error('Outer stack:')
-                    for item in traceback.format_list(outer_stack):
-                        self.log.error(item)
-                    self.log.error('Exception will be re-raised at next call.')
+
+                    if self._outer_stack:
+                        self.log.error('Outer stack:')
+                        for item in traceback.format_list(outer_stack):
+                            self.log.error(item)
+                        self.log.error('Exception will be re-raised at next call.')
+                    else:
+                        self.log.error(
+                            'Unable to show the outer stack. Pass '
+                            'outer_stack=True when initializing the '
+                            'SqliteDict instance to show the outer stack.'
+                        )
 
                 if res:
                     for rec in cursor:
@@ -483,13 +502,17 @@ class SqliteMultithread(Thread):
         """
         `execute` calls are non-blocking: just queue up the request and return immediately.
         """
+        self._wait_for_initialization()
         self.check_raise_error()
+        stack = None
+
+        if self._outer_stack:
+            # NOTE: This might be a lot of information to pump into an input
+            # queue, affecting performance.  I've also seen earlier versions of
+            # jython take a severe performance impact for throwing exceptions
+            # so often.
+            stack = traceback.extract_stack()[:-1]
 
-        # NOTE: This might be a lot of information to pump into an input
-        # queue, affecting performance.  I've also seen earlier versions of
-        # jython take a severe performance impact for throwing exceptions
-        # so often.
-        stack = traceback.extract_stack()[:-1]
         self.reqs.put((req, arg or tuple(), res, stack))
 
     def executemany(self, req, items):
@@ -546,4 +569,30 @@ class SqliteMultithread(Thread):
             # returning (by semaphore '--no more--'
             self.select_one('--close--')
             self.join()
-#endclass SqliteMultithread
+
+    def _wait_for_initialization(self):
+        """
+        Polls the 'initialized' flag to be set by the started Thread in run().
+        """
+        # A race condition may occur without waiting for initialization:
+        # __init__() finishes with the start() call, but the Thread needs some time to actually start working.
+        # If opening the database file fails in run(), an exception will occur and self.exception will be set.
+        # But if we run check_raise_error() before run() had a chance to set self.exception, it will report
+        # a false negative: An exception occured and the thread terminates but self.exception is unset.
+        # This leads to a deadlock while waiting for the results of execute().
+        # By waiting for the Thread to set the initialized flag, we can ensure the thread has successfully
+        # opened the file - and possibly set self.exception to be detected by check_raise_error().
+
+        start_time = time.time()
+        while time.time() - start_time < self.timeout:
+            if self._sqlitedict_thread_initialized or self.exception:
+                return
+            time.sleep(0.1)
+        raise TimeoutError("SqliteMultithread failed to flag initialization within %0.0f seconds." % self.timeout)
+
+
+#
+# This is here for .github/workflows/release.yml
+#
+if __name__ == '__main__':
+    print(__version__)
